/// <reference types="node" />
import { BankInfo, SlotType, IPosition, TradeSlotType, SlotInfo, StatusInfo } from "./definitions/adventureland.js";
import { Attribute, BankPackName, CharacterType, CXData, DamageType, EmotionName, GData, ItemName, MapName, MonsterName, NPCName, SkillName } from "./definitions/adventureland-data.js";
import { AchievementProgressData, CharacterData, ServerData, ChestOpenedData, ChestData, EntitiesData, EvalData, GameResponseData, NewMapData, PartyData, QData, TrackerData, PlayersData, ItemData, ItemDataTrade, PlayerData, FriendData } from "./definitions/adventureland-server.js";
import { PathfinderOptions } from "./definitions/pathfinder.js";
import { Entity } from "./Entity.js";
import { Observer } from "./Observer.js";
import { Player } from "./Player.js";
export declare class Character extends Observer implements CharacterData {
    protected userAuth: string;
    protected characterID: string;
    timeouts: Map<string, NodeJS.Timeout>;
    achievements: Map<string, AchievementProgressData>;
    bank: BankInfo;
    chests: Map<string, ChestData>;
    nextSkill: Map<SkillName, Date>;
    partyData: PartyData;
    ready: boolean;
    afk: "code";
    age: number;
    apiercing: number;
    blast: number;
    controller: string;
    in: string;
    name: string;
    id: string;
    ctype: CharacterType;
    abs: boolean;
    angle: number;
    armor: number;
    attack: number;
    c: any;
    cid: number;
    cx: CXData;
    damage_type: DamageType;
    focus?: string;
    frequency: number;
    going_x: number;
    going_y: number;
    gold: number;
    hp: number;
    level: number;
    m: number;
    mp_cost: number;
    max_hp: number;
    max_mp: number;
    move_num: number;
    moving: boolean;
    mp: number;
    npc?: NPCName;
    owner: string;
    party?: string;
    pdps: number;
    q: {
        upgrade?: {
            len: number;
            ms: number;
            num: number;
        };
        compound?: {
            len: number;
            ms: number;
            num: number;
            nums: number[];
        };
        exchange?: {
            len: number;
            ms: number;
            num: number;
            name: ItemName;
            id: ItemName;
            q: number;
        };
    };
    range: number;
    resistance: number;
    rip: boolean;
    rpiercing: number;
    s: StatusInfo;
    skin: string;
    slots: SlotInfo;
    speed: number;
    stand?: boolean | "cstand" | "stand0";
    tp: boolean;
    emx: {
        [T in EmotionName]?: number;
    };
    explosion: number;
    firesistance: number;
    fzresistance: number;
    mp_reduction: number;
    pnresistance: number;
    stun: number;
    int: number;
    str: number;
    dex: number;
    vit: number;
    for: number;
    max_xp: number;
    goldm: number;
    xpm: number;
    xp: number;
    luckm: number;
    isize: number;
    esize: number;
    cash: number;
    targets: number;
    target?: string;
    evasion: number;
    miss: number;
    reflection: number;
    lifesteal: number;
    manasteal: number;
    crit: number;
    critdamage: number;
    dreturn: number;
    tax: number;
    xrange: number;
    items: ItemData[];
    cc: number;
    ipass?: string;
    friends?: string[];
    acx?: any;
    xcx?: string[];
    hitchhikers?: [string, any][];
    user?: BankInfo;
    fear: number;
    courage: number;
    mcourage: number;
    pcourage: number;
    constructor(userID: string, userAuth: string, characterID: string, g: GData, serverData: ServerData);
    protected updateLoop(): Promise<void>;
    parseCharacter(data: CharacterData | PlayerData): void;
    protected parseEntities(data: EntitiesData): void;
    protected parseEval(data: EvalData): void;
    protected parseGameResponse(data: GameResponseData): void;
    protected parseNewMap(data: NewMapData): void;
    protected parseQData(data: QData): void;
    protected setNextSkill(skill: SkillName, next: Date): void;
    protected updatePositions(): void;
    /**
     * TODO: Add fail check for logging in with too many characters to one server
     *
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    connect(): Promise<void>;
    disconnect(): void;
    /**
     * This function will request all nearby entities and players from the server. You can use it to make sure we have the latest data.
     * NOTE: There is a rather high code call cost to this, don't call it too often.
     */
    requestEntitiesData(): Promise<EntitiesData>;
    /**
     * This function is a hack to get the server to respond with a player data update. It will respond with two...
     */
    requestPlayerData(): Promise<CharacterData>;
    /**
     * Accepts a friend request.
     *
     * @param {string} id
     * @return {*}  {Promise<FriendData>}
     * @memberof Character
     */
    acceptFriendRequest(id: string): Promise<FriendData>;
    /**
     * Accepts a magiport request from another character
     * @param name ID of the character that offered a magiport.
     */
    acceptMagiport(name: string): Promise<IPosition>;
    /**
     * Accepts another character's party invite.
     * @param id The ID of the character's party you want to accept the invite for.
     */
    acceptPartyInvite(id: string): Promise<PartyData>;
    acceptPartyRequest(id: string): Promise<PartyData>;
    /**
     * NOTE: We can't name this function `attack` because of the property `attack` that specifies damage.s
     * @param id The ID of the entity or player to attack
     */
    basicAttack(id: string): Promise<string>;
    buy(itemName: ItemName, quantity?: number): Promise<number>;
    buyFromMerchant(id: string, slot: TradeSlotType, rid: string, quantity?: number): Promise<ItemData>;
    /**
     * Buys an item from Ponty. Get items from `getPontyItems()`
     *
     * @param {ItemDataTrade} item
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    buyFromPonty(item: ItemDataTrade): Promise<void>;
    /**
     * Calculates the type of targets attacking you.
     *
     * The first element is the current number of targets of the given damage type.
     *
     * The second element is our character's courage
     *
     * @return {*}  {{
     *         magical: [number, number];
     *         physical: [number, number];
     *         pure: [number, number];
     *         }}
     * @memberof Character
     */
    calculateTargets(): {
        magical: number;
        physical: number;
        pure: number;
    };
    calculateDamageRange(defender: Character | Entity | Player, skill?: SkillName): [number, number];
    /**
     * Returns the *minimum* gold required to obtain the given item.
     *
     * @param {ItemData} item - The item to calculate the minimum cost for
     * @return {*}  {number} - The cost of the item
     * @memberof Character
     */
    calculateItemCost(item: ItemData): number;
    calculateItemGrade(item: ItemData): number;
    canBuy(item: ItemName, options?: {
        ignoreLocation?: boolean;
    }): boolean;
    /**
     * Returns true if you have the required items and gold to craft the item, and you
     * are near the NPC where you can craft this item.
     *
     * @param {ItemName} itemToCraft
     * @return {*}  {boolean}
     * @memberof Character
     */
    canCraft(itemToCraft: ItemName, options?: {
        ignoreLocation?: boolean;
    }): boolean;
    /**
     * Returns true if you have enough of the exchangeable items, and you are
     * near the NPC where you can exchange this item.
     *
     * @param {ItemName} itemToExchange
     * @param {{
     *         ignoreLocation?: boolean
     *     }} [options]
     * @return {*}  {boolean}
     * @memberof Character
     */
    canExchange(itemToExchange: ItemName, options?: {
        ignoreLocation?: boolean;
    }): boolean;
    /**
     * Returns true if we can 100% kill the entity in one shot.
     *
     * @param {Entity} entity
     * @param {SkillName} [skill="attack"]
     * @return {*}  {boolean}
     * @memberof Character
     */
    canKillInOneShot(entity: Entity, skill?: SkillName): boolean;
    /**
     * Returns true if we can sell items from where we are standing
     *
     * @return {*}  {boolean}
     * @memberof Character
     */
    canSell(): boolean;
    /**
     * UNFINISHED. DO NOT USE YET.
     * TODO: Finish
     *
     * @param {number} itemPos
     * @param {number} scrollPos
     * @param {number} [offeringPos]
     * @return {*}  {boolean}
     * @memberof Character
     */
    canUpgrade(itemPos: number, scrollPos: number, offeringPos?: number): boolean;
    /**
     * Returns true if we can use the skill.
     * This function checks various requirements, such as level, having a required item equipped, etc.
     * If you only want to check the cooldown, use isOnCooldown(skill) or getCooldown(skill).
     *
     * @param {SkillName} skill
     * @param {{
     *         ignoreCooldown?: boolean,
     *         ignoreEquipped?: boolean
     *     }} [options]
     * @return {*}  {boolean}
     * @memberof Character
     */
    canUse(skill: SkillName, options?: {
        ignoreCooldown?: boolean;
        ignoreEquipped?: boolean;
    }): boolean;
    compound(item1Pos: number, item2Pos: number, item3Pos: number, cscrollPos: number, offeringPos?: number): Promise<boolean>;
    /**
     * Crafts the given item name using items in the character's inventory.
     *
     * @param {ItemName} item
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    craft(item: ItemName): Promise<void>;
    depositGold(gold: number): unknown;
    /**
     * Deposits an item in your bank.
     *
     * @param {number} inventoryPos
     * @param {BankPackName} [bankPack]
     * @param {*} [bankSlot=-1]
     * @return {*}  {unknown}
     * @memberof Character
     */
    depositItem(inventoryPos: number, bankPack?: BankPackName, bankSlot?: number): Promise<unknown>;
    /**
     * Perform an emotion
     *
     * @param {EmotionName} emotionName
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    emote(emotionName: EmotionName): Promise<void>;
    /**
     * Use this function to enter dungeons.
     *
     * See `transport` for traveling through normal doors.
     *
     * @param {MapName} map
     * @param {string} [instance] Instance ID string. If not set, a key will be consumed from your inventory to create a new one.
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    enter(map: MapName, instance?: string): Promise<void>;
    equip(inventoryPos: number, equipSlot?: SlotType): Promise<void>;
    exchange(inventoryPos: number): Promise<void>;
    finishMonsterHuntQuest(): Promise<void>;
    /**
     * Returns a list of nearby entities, with optional filters
     *
     * @param {{
     *         canWalkTo?: boolean
     *         couldGiveCredit?: boolean
     *         withinRange?: number
     *         targetingMe?: boolean
     *         targetingPartyMember?: boolean
     *         targetingPlayer?: string
     *         type?: MonsterName
     *         typeList?: MonsterName[]
     *         level?: number
     *         levelGreaterThan?: number
     *         levelLessThan?: number
     *         willBurnToDeath?: boolean
     *         willDieToProjectiles?: boolean
     *     }} [filters={}]
     * @return {*}  {Entity[]}
     * @memberof Character
     */
    getEntities(filters?: {
        canWalkTo?: boolean;
        couldGiveCredit?: boolean;
        withinRange?: number;
        targetingMe?: boolean;
        targetingPartyMember?: boolean;
        targetingPlayer?: string;
        type?: MonsterName;
        typeList?: MonsterName[];
        level?: number;
        levelGreaterThan?: number;
        levelLessThan?: number;
        willBurnToDeath?: boolean;
        willDieToProjectiles?: boolean;
    }): Entity[];
    /**
     * Returns the index of the first empty slot in the given inventory
     *
     * @param {*} [items=this.items] The inventory to look in
     * @return {*}  {number} The index of the first empty slot
     * @memberof Character
     */
    getFirstEmptyInventorySlot(items?: ItemData[]): number;
    getMonsterHuntQuest(): Promise<void>;
    /**
     * Returns a list of the players that are online on the server
     *
     * @return {*}  {Promise<PlayersData>}
     * @memberof Character
     */
    getPlayers(): Promise<PlayersData>;
    getPontyItems(): Promise<ItemDataTrade[]>;
    /**
     * Retrieves the entity our character is currently targeting.
     *
     * @return {*}  {Entity}
     * @memberof Character
     */
    getTargetEntity(): Entity;
    /**
     * Retrieves tracker data
     *
     * @return {*}  {Promise<TrackerData>}
     * @memberof Character
     */
    getTrackerData(): Promise<TrackerData>;
    /**
     * Returns true if our inventory is full, false otherwise
     *
     * @return {*}  {boolean}
     * @memberof Character
     */
    isFull(): boolean;
    /**
     * Returns true if our character is scared, false otherwise
     *
     * @return {*}  {boolean}
     * @memberof Character
     */
    isScared(): boolean;
    /**
     * Removes a party member
     *
     * @param {string} toKick
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    kickPartyMember(toKick: string): Promise<void>;
    /**
     * For use on 'cyberland' and 'jail' to leave the map. You will be transported to the spawn on "main".
     *
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    leaveMap(): Promise<void>;
    leaveParty(): Promise<void>;
    /**
     * Moves the character to a given location. If the character can not move there safely,
     * i.e. there's a wall in the way, then we will move to the closest we can walk there in
     * a straight line.
     *
     * If you want this function to return after we complete the move, use `await`.
     *
     * If you start a new move before the last move is finished, the last move's promise will resolve.
     *
     * @param {number} x
     * @param {number} y
     * @param {{ disableSafetyCheck: boolean }} [options]
     *
     * disableSafetyCheck - If set to true, move() will not check map bounds
     *
     * @return {*}  {Promise<IPosition>}
     * @memberof Character
     */
    move(x: number, y: number, options?: {
        disableSafetyCheck: boolean;
    }): Promise<IPosition>;
    openChest(id: string): Promise<ChestOpenedData>;
    regenHP(): Promise<void>;
    regenMP(): Promise<void>;
    /**
     * If you are dead, you can call this function to respawn.
     *
     * @param {boolean} [safe] If set, you will spawn in Wizard's cave instead of near the goos.
     * @return {*}  {Promise<IPosition>}
     * @memberof Character
     */
    respawn(safe?: boolean): Promise<IPosition>;
    scare(): Promise<string[]>;
    sell(itemPos: number, quantity?: number): Promise<void>;
    sellToMerchant(id: string, slot: TradeSlotType, rid: string, q: number): Promise<void>;
    sendCM(to: string[], message: unknown): Promise<void>;
    sendFriendRequest(id: string): Promise<void>;
    sendGold(to: string, amount: number): Promise<number>;
    sendItem(to: string, inventoryPos: number, quantity?: number): Promise<void>;
    /**
     * Invites the given character to our party.
     * @param id The character ID to invite to our party.
     */
    sendPartyInvite(id: string): Promise<void>;
    /**
     * Requests to join another character's party.
     * @param id The character ID to request a party invite from.
     */
    sendPartyRequest(id: string): Promise<void>;
    /**
     * Shifts a booster to the given type
     *
     * @param {number} booster the inventory popsition of the booster
     * @param {("goldbooster" | "luckbooster" | "xpbooster")} to the type you want to shift it to
     * @memberof Character
     */
    shiftBooster(booster: number, to: "goldbooster" | "luckbooster" | "xpbooster"): Promise<void>;
    protected lastSmartMove: number;
    smartMoving: boolean;
    /**
     * Used to move long distances strategically, i.e. avoiding walking through walls.
     * You can use this function to move across maps.
     *
     * @param {(IPosition | ItemName | MapName | MonsterName | NPCName)} to
     * @param {PathfinderOptions} [options={
     *         avoidTownWarps: false,
     *         getWithin: 0,
     *         useBlink: false
     *     }]
     * @return {*}  {Promise<IPosition>}
     * @memberof Character
     */
    smartMove(to: IPosition | ItemName | MapName | MonsterName | NPCName, options?: PathfinderOptions): Promise<IPosition>;
    /**
     * Starts "Konami" mode.
     *
     * In Konami mode, you can only attack the monster it specifies.
     *
     * You have a (very low) chance to get a special item in this mode.
     *
     * To exit Konami mode, you need to disconnect and reconnect
     *
     * @return {*}  {Promise<MonsterName>} The type of monster you need to target
     * @memberof Character
     */
    startKonami(): Promise<MonsterName>;
    stopSmartMove(): Promise<IPosition>;
    stopWarpToTown(): Promise<void>;
    /**
     * Swaps two items in your inventory
     *
     * @param {number} itemPosA
     * @param {number} itemPosB
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    swapItems(itemPosA: number, itemPosB: number): Promise<void>;
    takeMailItem(mailID: string): Promise<void>;
    throwSnowball(target: string, snowball?: number): Promise<string>;
    /**
     * Use this function to travel through normal doors.
     *
     * See `enter` for entering dungeons.
     *
     * @param {MapName} map The map to move to
     * @param {number} spawn The spawn to move to for the given map
     * @return {*}  {Promise<void>}
     * @memberof Character
     */
    transport(map: MapName, spawn: number): Promise<void>;
    /**
     * Unequips the item in the given slot.
     * Also used to remove items from the merchant stand.
     *
     * @param {(SlotType | TradeSlotType)} slot
     * @return {*}  {Promise<number>} The position of the unequipped item in the inventory
     * @memberof Character
     */
    unequip(slot: SlotType | TradeSlotType): Promise<number>;
    /**
     * Unfriend another player.
     * NOTE: `data.name` may not equal `id`. The event uses the player's 1st character's name.
     *
     * @param {string} id
     * @return {*}  {Promise<FriendData>}
     * @memberof Character
     */
    unfriend(id: string): Promise<FriendData>;
    upgrade(itemPos: number, scrollPos: number, offeringPos?: number): Promise<boolean>;
    useHPPot(itemPos: number): Promise<void>;
    useMPPot(itemPos: number): Promise<void>;
    warpToJail(): Promise<IPosition>;
    warpToTown(): Promise<IPosition>;
    withdrawGold(gold: number): unknown;
    withdrawItem(bankPack: BankPackName, bankPos: number, inventoryPos?: number): Promise<unknown>;
    /**
     * Returns true if the entity has a >0% chance to die from projectiles already cast.
     *
     * @param {Map<string, ActionData>} projectiles (e.g.: bot.projectiles)
     * @param {Map<string, Player>} players (e.g.: bot.players)
     * @param {Map<string, Player>} entities (e.g.: bot.entitites)
     * @return {*}  {boolean}
     * @memberof Entity
     */
    couldDieToProjectiles(): boolean;
    /**
     * Returns the number of items that match the given parameters
     * @param itemName The item to look for
     * @param inventory Where to look for the item
     * @param filters Filters to help search for specific properties on items
     */
    countItem(item: ItemName, inventory?: ItemData[], filters?: {
        levelGreaterThan?: number;
        levelLessThan?: number;
    }): number;
    getCooldown(skill: SkillName): number;
    getNearestAttackablePlayer(): {
        player: Player;
        distance: number;
    };
    /**
     * Returns a boolean corresponding to whether or not we have a PvP marked item in our inventory
     * @param inv The inventory to look in
     */
    hasPvPMarkedItem(inv?: ItemData[]): boolean;
    /**
     * Returns a boolean corresponding to whether or not the item is in our inventory.
     * @param iN The item to look for
     * @param inv Where to look for the item
     */
    hasItem(iN: ItemName, inv?: ItemData[], filters?: {
        level?: number;
        levelGreaterThan?: number;
        levelLessThan?: number;
        locked?: boolean;
        quantityGreaterThan?: number;
        statType?: Attribute;
    }): boolean;
    /**
     * Returns a boolean corresponding to whether or not we have a given item equipped.
     * @param itemName The item to look for
     */
    isEquipped(itemName: ItemName): boolean;
    /**
     * Returns true if our skill is on cooldown, false otherwise.
     * This function does not check if you can use the skill.
     * If you want to check if you have all the requirements to use it, see canUse(skill).
     *
     * @param {SkillName} skill
     * @return {*}  {boolean}
     * @memberof Character
     */
    isOnCooldown(skill: SkillName): boolean;
    /**
     * Can we attack other players?
     *
     * @return {*}  {boolean}
     * @memberof Character
     */
    isPVP(): boolean;
    /**
     * Returns the index of the item in the given inventory
     * @param iN The item to look for
     * @param inv Where to look for the item
     * @param filters Filters to help search for specific properties on items
     */
    locateItem(iN: ItemName, inv?: ItemData[], filters?: {
        level?: number;
        levelGreaterThan?: number;
        levelLessThan?: number;
        locked?: boolean;
        quantityGreaterThan?: number;
        returnHighestLevel?: boolean;
        statType?: Attribute;
    }): number;
    /**
     * Returns a list of indexes of the items in the given inventory
     * @param iN The item to look for
     * @param inv Where to look for the item
     * @param filters Filters to help search for specific properties on items
     */
    locateItems(iN: ItemName, inv?: ItemData[], filters?: {
        level?: number;
        levelGreaterThan?: number;
        levelLessThan?: number;
        locked?: boolean;
        quantityGreaterThan?: number;
        statType?: Attribute;
    }): number[];
    /**
     * Returns an object with keys for each item name.
     * Each item object has keys for each level.
     * Each level object is an array of item positions.
     *
     * Set `minAmount` to 2 to find duplicate items. Set `minAmount` to 3 to find triplicate items.
     *
     * @param {*} [inventory=this.items]
     * @param {{ minAmount?: number }} [options]
     * @return {*}  {{
     *         [name in ItemName]?: {
     *             [level in number]?: number[];
     *         }
     *     }}
     * @memberof Character
     */
    locateItemsByLevel(inventory?: ItemData[], options?: {
        excludeLockedItems?: boolean;
        excludeSpecialItems?: boolean;
        minAmount?: number;
    }): {
        [name in ItemName]?: {
            [level in number]?: number[];
        };
    };
    locateMonster(mType: MonsterName): IPosition[];
    locateNPC(npcID: NPCName): IPosition[];
    locateCraftNPC(itemName: ItemName): IPosition;
    locateExchangeNPC(itemName: ItemName): IPosition;
}
