import { Constants } from "./Constants.js";
import { PingCompensatedCharacter } from "./PingCompensatedCharacter.js";
import { Tools } from "./Tools.js";
export class Merchant extends PingCompensatedCharacter {
    ctype = "merchant";
    closeMerchantStand() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [closeMerchantStand].");
        if (!this.stand)
            return Promise.resolve(); // It's already closed
        const closed = new Promise((resolve, reject) => {
            const checkStand = (data) => {
                if (!data.stand) {
                    this.socket.off("player", checkStand);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("player", checkStand);
                reject(`closeMerchantStand timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("player", checkStand);
        });
        this.socket.emit("merchant", { close: 1 });
        return closed;
    }
    /**
     * Fish for items
     *
     * @return {*}  {Promise<void>}
     * @memberof Merchant
     */
    fish() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [fish].");
        let startedFishing = false;
        if (this.c.fishing)
            startedFishing = true; // We're already fishing!?
        const fished = new Promise((resolve, reject) => {
            const caughtCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]fishing['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    resolve();
                }
            };
            const failCheck1 = (data) => {
                if (typeof data == "string") {
                    if (data == "skill_cant_wtype") {
                        this.socket.off("game_response", failCheck1);
                        this.socket.off("ui", failCheck2);
                        this.socket.off("eval", caughtCheck);
                        this.socket.off("player", failCheck3);
                        reject("We don't have a fishing rod equipped");
                    }
                }
                else if (typeof data == "object") {
                    if (data.response == "cooldown" && data.place == "fishing" && data.skill == "fishing") {
                        this.socket.off("game_response", failCheck1);
                        this.socket.off("ui", failCheck2);
                        this.socket.off("eval", caughtCheck);
                        this.socket.off("player", failCheck3);
                        reject(`Fishing is on cooldown (${data.ms}ms remaining)`);
                    }
                }
            };
            const failCheck2 = (data) => {
                if (data.type == "fishing_fail" && data.name == this.id) {
                    // NOTE: We might not be in a fishing area?
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    reject("We failed to fish.");
                }
                else if (data.type == "fishing_none") {
                    // We fished, but we didn't catch anything
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    resolve();
                }
            };
            const failCheck3 = (data) => {
                if (!startedFishing && data.c.fishing) {
                    startedFishing = true;
                }
                else if (startedFishing && !data.c.fishing) {
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    // TODO: Is there a reliable way to figure out if we got interrupted?
                    // TODO: Maybe the eval cooldown?
                    resolve(); // We fished and caught nothing, or got interrupted.
                }
            };
            setTimeout(() => {
                this.socket.off("game_response", failCheck1);
                this.socket.off("ui", failCheck2);
                this.socket.off("eval", caughtCheck);
                this.socket.off("player", failCheck3);
                reject("fish timeout (20000ms)");
            }, 20000);
            this.socket.on("game_response", failCheck1);
            this.socket.on("eval", caughtCheck);
            this.socket.on("ui", failCheck2);
            this.socket.on("player", failCheck3);
        });
        this.socket.emit("skill", { name: "fishing" });
        return fished;
    }
    // TODO: Add promises
    joinGiveaway(slot, id, rid) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [joinGiveaway].");
        const merchant = this.players.get(id);
        if (!merchant || Tools.distance(this, merchant) > Constants.NPC_INTERACTION_DISTANCE)
            return Promise.reject(`${id} is too far away.`);
        if (!merchant.slots[slot]?.giveaway)
            return Promise.reject(`${id}'s slot ${slot} is not a giveaway.`);
        if (merchant.slots[slot]?.list.includes(id))
            return Promise.resolve(); // We've already joined it
        this.socket.emit("join_giveaway", { slot: slot, id: id, rid: rid });
    }
    /**
     * Lists an item for sale on your merchant stand
     *
     * @param {number} itemPos
     * @param {TradeSlotType} tradeSlot
     * @param {number} price
     * @param {number} [quantity=1]
     * @return {*}  {unknown}
     * @memberof Merchant
     */
    async listForSale(itemPos, tradeSlot, price, quantity = 1) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [listForSale].");
        const itemInfo = this.items[itemPos];
        if (!itemInfo)
            return Promise.reject(`We do not have an item in slot ${itemPos}`);
        if (price <= 0)
            return Promise.reject("The lowest you can set the price is 1.");
        const slotInfo = this.slots[tradeSlot];
        if (slotInfo) {
            const gInfo = this.G.items[itemInfo.name];
            if (itemInfo.name == slotInfo.name // Same item
                && price >= slotInfo.price // Same, or higher price
                && gInfo.s && (quantity + slotInfo.q) <= gInfo.s) // Stackable
             {
                if (itemPos !== 0) {
                    // Swap items so when it gets stacked, it gets stacked in the correct position
                    await this.swapItems(0, itemPos);
                }
                // Unequip, so we can combine the two slots
                await this.unequip(tradeSlot);
                quantity += slotInfo.q;
                if (itemPos !== 0) {
                    // Swap back
                    await this.swapItems(0, itemPos);
                }
            }
            else {
                return Promise.reject(`We are already trading something in ${tradeSlot}.`);
            }
        }
        const listed = new Promise((resolve, reject) => {
            const failCheck1 = (data) => {
                if (typeof data == "string") {
                    if (data == "slot_occupied") {
                        this.socket.off("game_response", failCheck1);
                        this.socket.off("disappearing_text", failCheck2);
                        this.socket.off("player", successCheck);
                        reject(`We are already listing something in ${tradeSlot}.`);
                    }
                }
            };
            const failCheck2 = (data) => {
                if (data.message == "CAN'T EQUIP" && data.id == this.id) {
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("disappearing_text", failCheck2);
                    this.socket.off("player", successCheck);
                    reject(`We failed listing the item in ${tradeSlot}.`);
                }
            };
            const successCheck = (data) => {
                const newTradeSlot = data.slots[tradeSlot];
                if (newTradeSlot && newTradeSlot.name == itemInfo.name && newTradeSlot.q == quantity) {
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("disappearing_text", failCheck2);
                    this.socket.off("player", successCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("game_response", failCheck1);
                this.socket.off("disappearing_text", failCheck2);
                this.socket.off("player", successCheck);
                reject("listForSale timeout (1000ms)");
            }, 1000);
            this.socket.on("game_response", failCheck1);
            this.socket.on("disappearing_text", failCheck2);
            this.socket.on("player", successCheck);
        });
        this.socket.emit("equip", {
            num: itemPos,
            price: price,
            q: quantity,
            slot: tradeSlot
        });
        return listed;
    }
    // TODO: Add promises
    merchantCourage() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [merchantCourage].");
        this.socket.emit("skill", { name: "mcourage" });
    }
    mine() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [mine].");
        let startedMining = false;
        if (this.c.mining)
            startedMining = true; // We're already mining!?
        const mined = new Promise((resolve, reject) => {
            const caughtCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]mining['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    resolve();
                }
            };
            const failCheck1 = (data) => {
                if (typeof data == "string") {
                    if (data == "skill_cant_wtype") {
                        this.socket.off("game_response", failCheck1);
                        this.socket.off("ui", failCheck2);
                        this.socket.off("eval", caughtCheck);
                        this.socket.off("player", failCheck3);
                        reject("We don't have a pickaxe equipped");
                    }
                }
                else if (typeof data == "object") {
                    if (data.response == "cooldown" && data.place == "mining" && data.skill == "mining") {
                        this.socket.off("game_response", failCheck1);
                        this.socket.off("ui", failCheck2);
                        this.socket.off("eval", caughtCheck);
                        this.socket.off("player", failCheck3);
                        reject(`Mining is on cooldown (${data.ms}ms remaining)`);
                    }
                }
            };
            const failCheck2 = (data) => {
                if (data.type == "mining_fail" && data.name == this.id) {
                    // NOTE: We might not be in a mining area?
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    reject("We failed to mine.");
                }
                else if (data.type == "mining_none") {
                    // We mined, but we didn't get anything
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    resolve();
                }
            };
            const failCheck3 = (data) => {
                if (!startedMining && data.c.mining) {
                    startedMining = true;
                }
                else if (startedMining && !data.c.mining) {
                    this.socket.off("game_response", failCheck1);
                    this.socket.off("ui", failCheck2);
                    this.socket.off("eval", caughtCheck);
                    this.socket.off("player", failCheck3);
                    // TODO: Is there a reliable way to figure out if we got interrupted?
                    // TODO: Maybe the eval cooldown?
                    resolve(); // We mined and caught nothing, or got interrupted.
                }
            };
            setTimeout(() => {
                this.socket.off("game_response", failCheck1);
                this.socket.off("ui", failCheck2);
                this.socket.off("eval", caughtCheck);
                this.socket.off("player", failCheck3);
                reject("mine timeout (20000ms)");
            }, 20000);
            this.socket.on("game_response", failCheck1);
            this.socket.on("eval", caughtCheck);
            this.socket.on("ui", failCheck2);
            this.socket.on("player", failCheck3);
        });
        this.socket.emit("skill", { name: "mining" });
        return mined;
    }
    mluck(target) {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [mluck].");
        let previousMs = 0;
        if (target !== this.id) {
            const player = this.players.get(target);
            if (!player)
                return Promise.reject(`Could not find ${target} to mluck.`);
            if (player.npc)
                return Promise.reject(`${target} is an NPC. You can't mluck NPCs.`);
            if (player.s.mluck && player.s.mluck.strong // They have a strong mluck
                && ((player.owner && player.owner !== this.owner) // If they are public, check if they are from different owners
                    || (player.s.mluck.f !== this.id))) { // Else, rely on the character id
                return Promise.reject(`${target} has a strong mluck from ${player.s.mluck.f}.`);
            }
            previousMs = player.s?.mluck?.ms ?? 0;
        }
        else {
            previousMs = this.s?.mluck?.ms ?? 0;
        }
        const mlucked = new Promise((resolve, reject) => {
            const cooldownCheck = (data) => {
                if (/skill_timeout\s*\(\s*['"]mluck['"]\s*,?\s*(\d+\.?\d+?)?\s*\)/.test(data.code)) {
                    this.socket.off("entities", mluckCheck);
                    this.socket.off("game_response", failCheck);
                    this.socket.off("player", selfMluckCheck);
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            const mluckCheck = (data) => {
                for (const player of data.players) {
                    if (player.id == target
                        && player.s?.mluck?.f == this.id
                        && player.s?.mluck?.ms >= previousMs) {
                        this.socket.off("entities", mluckCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("player", selfMluckCheck);
                        this.socket.off("eval", cooldownCheck);
                        resolve();
                        return;
                    }
                }
            };
            const selfMluckCheck = (data) => {
                if (this.id == target
                    && data.s?.mluck?.f == this.id
                    && data.s?.mluck?.ms >= previousMs) {
                    this.socket.off("entities", mluckCheck);
                    this.socket.off("game_response", failCheck);
                    this.socket.off("player", selfMluckCheck);
                    this.socket.off("eval", cooldownCheck);
                    resolve();
                }
            };
            const failCheck = async (data) => {
                if (typeof data == "string") {
                    if (data == "skill_too_far") {
                        this.socket.off("entities", mluckCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("player", selfMluckCheck);
                        this.socket.off("eval", cooldownCheck);
                        await this.requestPlayerData().catch((e) => { console.error(e); });
                        reject(`We are too far from ${target} to mluck.`);
                    }
                    else if (data == "no_level") {
                        this.socket.off("entities", mluckCheck);
                        this.socket.off("game_response", failCheck);
                        this.socket.off("player", selfMluckCheck);
                        this.socket.off("eval", cooldownCheck);
                        reject("We aren't a high enough level to use mluck.");
                    }
                }
            };
            setTimeout(() => {
                this.socket.off("entities", mluckCheck);
                this.socket.off("game_response", failCheck);
                this.socket.off("player", selfMluckCheck);
                this.socket.off("eval", cooldownCheck);
                reject(`mluck timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("game_response", failCheck);
            this.socket.on("player", selfMluckCheck);
            this.socket.on("entities", mluckCheck);
            this.socket.on("eval", cooldownCheck);
        });
        this.socket.emit("skill", { id: target, name: "mluck" });
        return mlucked;
    }
    openMerchantStand() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [openMerchantStand].");
        if (this.stand)
            return Promise.resolve(); // It's already open
        // Find a suitable stand
        let stand;
        for (const item of ["supercomputer", "computer", "stand1", "stand0"]) {
            stand = this.locateItem(item);
            if (stand !== undefined)
                break;
        }
        if (stand == undefined)
            return Promise.reject("Could not find a suitable merchant stand in inventory.");
        const opened = new Promise((resolve, reject) => {
            const checkStand = (data) => {
                if (data.stand) {
                    this.socket.off("player", checkStand);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("player", checkStand);
                reject(`openMerchantStand timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("player", checkStand);
        });
        this.socket.emit("merchant", { num: stand });
        return opened;
    }
    massProduction() {
        if (!this.ready)
            return Promise.reject("We aren't ready yet [massProduction].");
        const massProductioned = new Promise((resolve, reject) => {
            const productedCheck = (data) => {
                if (data.type == "massproduction" && data.name == this.id) {
                    this.socket.off("ui", productedCheck);
                    resolve();
                }
            };
            setTimeout(() => {
                this.socket.off("ui", productedCheck);
                reject(`massProduction timeout (${Constants.TIMEOUT}ms)`);
            }, Constants.TIMEOUT);
            this.socket.on("ui", productedCheck);
        });
        this.socket.emit("skill", { name: "massproduction" });
        return massProductioned;
    }
}
//# sourceMappingURL=Merchant.js.map